/*
	---------------------------------------------------------------------------
		Relativity
		Adds relative positioning to GUIs created by AutoGUI's GUI designer
		Version: 1.2
		Date: 2020-05-26
		Author: boiler
		Requires AutoHotkey v1.1.31+
	---------------------------------------------------------------------------
	Changelog:
	v1.0	2020-02-03	Initial release
	v1.1	2020-02-12	Added Align/Nudge feature
	v1.2	2020-05-26	Multiple changes:
		- Added new Open File feature
		- Welcome screen display now a selectable option
		- Optionally remembers window position
		- Added menu with File, Script, Options, Help
		- Save As/Save always active allowing simple reordering to be saved
		- Uses all local icons instead of calling them from the Windows files
		- Button ribbon updated/improved
	---------------------------------------------------------------------------
	Acknowledgements:
	- Alguimist for the AutoGUI editor and GUI design tool:
	  https://www.autohotkey.com/boards/viewtopic.php?f=64&t=40290	
	- lexikos for "How to: Run Dynamic Script... Through a Pipe!" and...
	- HotKeyIt for his RunTempScript function using the above, both found here:
	  https://autohotkey.com/board/topic/23575-how-to-run-dynamic-script-through-a-pipe
	- just me for GuiControlTips:
	  https://www.autohotkey.com/boards/viewtopic.php?f=6&t=2598
	- FanaticGuru for GuiButtonIcon:
	  https://www.autohotkey.com/boards/viewtopic.php?t=1985
	- just me for the Rich Text Format control tutorial:
	  https://www.autohotkey.com/boards/viewtopic.php?f=11&t=44874
	  and RTF control tutorial translation from German by burque505:
	  https://www.autohotkey.com/boards/viewtopic.php?t=45481
*/

#Include, includes\RunTempScript.ahk
#Include, includes\GUIControlTips.ahk
#Include, includes\GuiButtonIcon.ahk
#Include, includes\GetTomDoc.ahk
#SingleInstance, Ignore
SetTitleMatchMode, 2
CoordMode, ToolTip, Client
Version := 1.1
global Control := []
global Testing := 1
global AlignBase, AlignDelta, AlignIndex, AlignSelectedRows, AlignText, AltTestGuiID, CurrAlignVal, GuiScriptFile, ScriptText, StartAlignValue, MainGuiHwnd, MatrixChanged, OutputScriptFile, OutputScriptText, TestGuiID, UserGuiHwnd, Version
MatrixChanged := -1
IniRead, ShowWelcome, settings.ini, Settings, ShowWelcome, 1
IniRead, RememberPos, settings.ini, Settings, RememberPos, 1
IniRead, Reloading, settings.ini, Status, Reloading, 0
Menu, Tray, Icon, icons\blackhole.ico
Menu, Tray, NoStandard
Menu, Tray, Add, Exit Relativity, ExitRelativity
Menu, Tray, Icon, 1&, icons\exit.ico, 1
RE_Dll := DllCall("LoadLibrary", "Str", "Msftedit.dll", "Ptr")
if ShowWelcome && !Reloading
	IntroGui()
IniWrite, 0, settings.ini, Status, Reloading

if Reloading
{
	IniRead, GuiScriptFile, settings.ini, Status, ReloadFile
	FileRead, ScriptText, % GuiScriptFile
}
else
{
	loop
	{
		FileSelectFile, GuiScriptFile, 1,, % "Select GUI script file", *.ahk
		if !GuiScriptFile
		{
			SplashMessage("No file selected`n`nExiting Relativity")
			ExitApp
		}
		FileRead, ScriptText, % GuiScriptFile
	} until ScriptIsValid(ScriptText)
}

loop, Parse, ScriptText, `n, `r
	if RegExMatch(A_LoopField, "^Gui Add, \w+, x\d+")
		Control.Push(new ControlDef(A_LoopField, A_Index))
gosub, MainGui
return

; ---------------------------------------
; ----  End of auto-execute section  ----
; ---------------------------------------

Class ControlDef
{
	__New(commandLine, lineNum)
	{
		RegExMatch(commandLine, "(^Gui Add), (\w+),\s(x[^,]*),{0,1}(.*)", m)
		this.Command := m1
		this.Type := m2
		this.Options := m3
		this.Text := Trim(m4)
		this.Line := lineNum
	}
}

GetSelectedRows()
{
	selected := []
	row := 0
	loop
	{
		row := LV_GetNext(row)
		if !row
			break
		selected.Push(row)
	}
	return selected
}

MoveRow(rowFrom, rowTo)
{
	loop, 10
		LV_GetText(t%A_Index%, rowFrom, A_Index)
	LV_Delete(rowFrom)
	LV_Insert(rowTo, "Select Icon" . (t10 ? t10 : 5), t1, t2, t3, t4, t5, t6, t7, t8, t9, t10)
}

RemoveToolbar(script)
{
	script := StrReplace(script, "#Include", "; #Include")
	script := StrReplace(script, "hToolbar := CreateToolbar()", "; hToolbar := CreateToolbar()")
	script := StrReplace(script, "CreateToolbar() {", "/*`nCreateToolbar() {")
	script .= "`n*/"
	return script
}

ScriptIsValid(script)
{
	goodSource := 0
	goodControlFound := 0
	badControlFound := 0
	loop, Parse, script, `n, `r
	{
		if (A_Index = 1) && (RegExMatch(A_LoopField, "^; Generated by AutoGUI"))
			goodSource := 1
		if RegExMatch(A_LoopField, "^Gui Add, \w+, x\d+ y\d+ w\d+ h\d+[^,]*,{0,1}.*")
			goodControlFound := 1
		if RegExMatch(A_LoopField, "^Gui Add, \w+, (xs|ys|xp|yp|wp|hp)[^,]*,{0,1}.*")
			badControlFound := 1
	}
	if (!goodSource || !goodControlFound || badControlFound)
	{
		MsgBox, 8257, No valid controls found, % "The selected file does not appear be generated directly by AutoGUI's GUI Designer and/or does not contain absoulte-positioned controls.`n`nPlease select a valid input GUI script file."
		IfMsgBox, Cancel
			ExitApp
		return false
	}
	return true
}

SplashMessage(message, time:=2, w:=200, h:=110, testMessage := 0)
{
	SplashTextOn, w, h, Relativity, % "`n" message
	loop, % 80 * time
	{
		Sleep, 10
		if (testMessage && !Testing)
		{
			SplashTextOff
			Gui, Show
			return
		}
	}
	SplashTextOff
}

Fontalyze(script, group)
{
	if group
		script := GroupFonts(script)
	lines := StrSplit(script, "`n", "`r")
	index := 1
	out := ""
	loop
	{
		if !(RegExMatch(lines[index], "^Gui Font,"))
			out .= lines[index++] "`n"
		else
		{
			matches := 0
			loop
			{
				if (lines[index] = lines[index + 3 * (A_Index - 1) + 3])
					matches++
				else
					break
			}
			if !matches
			{
				loop, 3
					out .= lines[index + A_Index - 1] "`n"
				index += 3
				continue
			}
			out .= lines[index] "`n"
			loop, % matches + 1
				out .= lines[index + 3 * (A_Index - 1) + 1] "`n"
			out .= lines[index + 2] "`n" ; any Gui Font line will do
			index += (matches + 1) * 3
		}
	} until index > lines.Count()
	return out
}

GroupFonts(script)
{
	lines := StrSplit(script, "`n", "`r")
	textoutput := ""
	candidates := []
	fontDefCounts := {}
	qualified := []
	lineRow := {}
	; map line numbers for controls to row numbers:
	thisRow := 1
	loop, % lines.Count()
	{
		if RegExMatch(lines[A_Index], "^Gui Add, \w+, x")
		{
			lineRow[A_Index] := thisRow ; lineRow[line] := row
			thisRow++
		}
	}
	; check that all the conditions are true to make a line a candidate for grouping:
	loop, % lines.Count()
	{
		conditions := 0
		conditions += !!RegExMatch(lines[A_Index], "^Gui Add,") ; condition 1
		conditions += !!RegExMatch(lines[A_Index - 1], "^Gui Font,") ; condition 2
		RegExMatch(lines[A_Index], "^Gui Add, \w+, \K.*?(?=,)", options)
		if options not contains Section,xs,ys,xp,yp,wp,hp ; condition 3
			conditions += 1
		if (RegExMatch(lines[A_Index + 2], "^Gui Add,"))
		{
			RegExMatch(lines[A_Index + 2], "^Gui Add, \w+, \K.*?(?=,)", options)
			if options not contains xp,yp,wp,hp ; condition 4 (possibility 1)
				conditions += 1
		}
		else if (RegExMatch(lines[A_Index + 3], "^Gui Add,"))
		{
			RegExMatch(lines[A_Index + 3], "^Gui Add, \w+, \K.*?(?=,)", options)
			if options not contains xp,yp,wp,hp ; condition 4 (possibility 2)
				conditions += 1
		}
		else if (Trim(lines[A_Index + 3]) = "") ; condition 4 (possibility 3)
			conditions += 1
		if (conditions = 4)
			candidates.Push(A_Index)
	}
	; check among candidates for multiple identical font definitions:
	loop, % candidates.Count()
	{
		if (fontDefCounts[lines[candidates[A_Index] - 1]] = "")
			fontDefCounts[lines[candidates[A_Index] - 1]] := 1
		else
			fontDefCounts[lines[candidates[A_Index] - 1]] += 1
	}
	for def, count in fontDefCounts
		if (count > 1)
			qualified.Push(def)
	for k, def in qualified
	{
		moveText .= def "`n"
		for k, line in candidates
			if (lines[line - 1] := def)
			{
				moveText .= lines[line] "`n"
				movedLines .= line - 1 "," line "," line + 1 ","
			}
		moveText .= "Gui Font`n"
	}
	movedLines := SubStr(movedLines, 1, -1) ; remove last comma
	; add lines from top of script:
	loop
	{
		if (SubStr(lines[A_Index], 1, 3) = "Gui")
		{
			controlsStart := A_Index
			break
		}
		newScript .= lines[A_Index] "`n"
	}
	; add lines that aren't moved:
	current := controlsStart
	loop
	{
		if (Trim(lines[current]) = "")
			break
		if current not in %movedLines%
			newScript .= lines[current] "`n"
		current++
	}
	; add lines that are moved:
	current := controlsStart
	loop
	{
		if (Trim(lines[current]) = "")
			break
		if current in %movedLines%
			newScript .= lines[current] "`n"
		current++
	}
	; add remaining lines:
	loop
	{
		newScript .= lines[current] "`n"
		current++
	} until current > lines.Count()
	; move rows in the ListView:
	currLine := 1
	rowsMoved := 0
	loop
	{
		if currLine in %movedLines%
		{
			currLine++ ; skip the font definition line
			loop, 10
				LV_GetText(t%A_Index%, lineRow[currLine] - rowsMoved, A_Index)
			LV_Delete(lineRow[currLine] - rowsMoved)
			LV_Add("Icon" . (t10 ? t10 : 5), t1, t2, t3, t4, t5, t6, t7, t8, t9, t10)
			currLine ++ ; skip over the end of font definition line	
			rowsMoved++
		}
		currLine++
	} until currLine > lines.Count()
	return newScript
}

AltTestScript(lineNum, element)
{
	loop, Parse, OutputScriptText, `n
	{
		if (A_Index = lineNum)
		{
			lineText := A_LoopField
			loop, 4
			{
				e := ["x", "y", "w", "h"][A_Index]
				if InStr(element, e)
					lineText := Shift5Pixels(lineText, e)
			}
			altScript .= lineText "`n"
		}
		else
			altScript .= A_LoopField "`n"
	}
	return altScript
}

Shift5Pixels(line, element)
{
	c := new ControlDef(line, 1)
	RegExMatch(c.Options, "\b" element "[sp]{0,1}\K[\+\-0-9]+", target)
	orig := target
	Switch SubStr(target, 1, 1)
	{
		Case "": target := "+5"
		Case "-":
			if (target < -5)
				target += 5
			else if (target = -5)
				target := ""
			else
				target := "+" target + 5
		Case "+": target := "+" target + 5
		Default: target += 5
	}
	if orig
		modOptions := RegExReplace(c.Options, "\b" element "[sp]{0,1}\K[\+\-0-9]+", target)
	else
		modOptions := RegExReplace(c.Options, "\b(" element "[sp])", "$1" target)

	return c.Command ", " c.Type ", " modOptions ", " c.Text
}

RunTestGuis(testScript, altTestScript)
{
	global AltTestGuiID, TestGuiID

	WinGet, preID, ID, A
	RunTempScript(RemoveToolbar(altTestScript))
	WinWaitNotActive, ahk_id %preID%
	WinGet, AltTestGuiID, ID, A
	RunTempScript(RemoveToolbar(testScript))
	WinWaitNotActive, ahk_id %AltTestGuiID%
	WinGet, TestGuiID, ID, A
	loop, 6
	{
		if !Testing
		{
			QuitTesting()
			return
		}
		WinActivate, ahk_id %TestGuiID%
		Sleep, 200
		WinActivate, ahk_id %AltTestGuiID%
		Sleep, 200
	}
	WinClose, ahk_id %TestGuiID%
	WinClose, ahk_id %AltTestGuiID%
}

QuitTesting()
{
	WinClose, ahk_id %AltTestGuiID%
	WinClose, ahk_id %TestGuiID%
	Gui, Show
}

EnableDisableUpdate()
{
	anchor := xs := ys := xp := yp := wp := hp := 0
	loop, % LV_GetCount()
	{
		LV_GetText(iconNum, A_Index, 10)
		if (iconNum = 2)
			anchor := 1
		LV_GetText(x, A_Index, 2)
		LV_GetText(y, A_Index, 3)
		LV_GetText(w, A_Index, 4)
		LV_GetText(h, A_Index, 5)
		xs := xs || (x = "s")
		ys := ys || (y = "s")
		xp := xp || (x = "p")
		yp := yp || (y = "p")
		wp := wp || (w = "p")
		hp := hp || (h = "p")
	}
	if (anchor && xs || xs) || (xp || yp || wp || hp) {
		GuiControl, Enable, UpdBt
		Menu, ScriptMenu, Enable, &Update Control Options
	} else {
		GuiControl, Disable, UpdBt
		Menu, ScriptMenu, Disable, &Update Control Options
	}
}

MainGui:
	Gui, +Resize +MinSize748x280 -MaximizeBox +HwndMainGuiHwnd +OwnDialogs
	Gui, Color, F6F6F6
	Menu, FileMenu, Add, &Open File, OpenFile
	Menu, FileMenu, Add, &Save As..., Save
	Menu, FileMenu, Add, &Run Gui Script, RunGuiScript
	Menu, FileMenu, Add, E&xit Relativity, ExitRelativity
	Menu, ScriptMenu, Add, &Update Control Options, Update
	Menu, ScriptMenu, Add, &Test GUI Changes, Test
	Menu, ScriptMenu, Add, &Fontalyzer, FontalyzerMenuSelect
	Menu, ScriptMenu, Add, &Fontalyzer Grouping, FontalyzerGroupMenuSelect
	Menu, OptionsMenu, Add, Show &Welcome Screen, WelcomeMenuSelect
	Menu, OptionsMenu, Add, &Remember Window Position, RememberPosMenuSelect
	Menu, HelpMenu, Add, User &Guide, Help
	Menu, HelpMenu, Add, &About Relativity, About
	Menu, MainMenu, Add, &File, :FileMenu
	Menu, MainMenu, Add, &Script, :ScriptMenu
	Menu, MainMenu, Add, &Options, :OptionsMenu
	Menu, MainMenu, Add, &Help, :HelpMenu
	Menu, ScriptMenu, Disable, &Update Control Options
	Menu, ScriptMenu, Disable, &Test GUI Changes
	Menu, ScriptMenu, Check, &Fontalyzer
	Menu, OptionsMenu, % (ShowWelcome ? "Check" : "Uncheck"), Show &Welcome Screen
	Menu, OptionsMenu, % (RememberPos ? "Check" : "Uncheck"), &Remember Window Position
	Gui, Menu, MainMenu
	Gui, Add, Button, xm y+8 w70 h25 gOpenFile HwndOpenHwnd Section, Open
	Gui, Add, Button, xp yp+30 wp hp vSvBt gSave HwndSaveHwnd, Save As
	Gui, Add, Button, xp yp+30 wp hp gRunGuiScript HwndRunHwnd, Run
	Gui, Add, Button, x+5 ys w70 h25 gHelp HwndHelpBtHwnd Section, Help
	Gui, Add, GroupBox, xp yp+24 wp-1 h61
	Gui, Add, Text, xp+6 yp+13 wp-10 hp-10 Center cGray BackgroundTrans, Hover over buttons to view tips
	Gui, Add, GroupBox, x+10 ys-5 w75 h90 c4040C0 Section, Align/Nudge
	Gui, Font, s12, Segoe UI
	Gui, Add, Button, xs+10 ys+20 w25 h28 gx HwndXHwnd
	Gui, Add, Button, x+5 yp wp hp gy HwndYHwnd
	Gui, Add, Button, xs+10 y+5 wp hp gw HwndWHwnd
	Gui, Add, Button, x+5 yp wp hp gh HwndHHwnd
	Gui, Font
	Gui, Add, GroupBox, xs+80 ys w155 h90 c4040C0 Section, Reorder Rows
	Gui, Add, Text, xs+10 ys+17, Shift/Ctrl to multi-select rows
	Gui, Add, Button, xs+10 ys+34 w23 h23 HwndUpBtHwnd gUpBtn
	Gui, Add, Button, xp y+3 wp hp HwndDownBtHwnd gDownBtn
	Gui, Add, Button, x+12 ys+34 wp hp HwndTopBtHwnd gTopBtn
	Gui, Add, Button, xp y+3 wp hp HwndBottomBtHwnd gBottomBtn
	Gui, Add, Button, x+12 ys+42 w65 h30 HwndGroupBtHwnd gGroupBtn, Group
	Gui, Add, GroupBox, xs+160 ys w100 h90 c4040C0 Section, Section Anchors
	Gui, Add, Button, xs+11 ys+20 w80 h28 gSelectAnchor hwndAnchorApplyHwnd, Section
	Gui, Add, Button, xp y+5 wp hp gClearAnchor hwndAnchorClearHwnd, Remove
	Gui, Add, GroupBox, xs+105 ys w155 h90 c4040C0 Section, Relative Controls
	Gui, Add, CheckBox, xs+15 ys+21 vXsBox gXsBox HwndXsHwnd, xs
	Gui, Add, CheckBox, xp yp+16 vYsBox gYsBox HwndYsHwnd, ys
	Gui, Add, CheckBox, xs+70 ys+21 vXpBox gXpBox HwndXpHwnd, xp
	Gui, Add, CheckBox, xp yp+16 vYpBox gYpBox HwndYpHwnd, yp
	Gui, Add, CheckBox, xs+110 ys+21 vWpBox gFocusLV HwndWpHwnd, wp
	Gui, Add, CheckBox, xp yp+16 vHpBox gFocusLV HwndHpHwnd, hp
	Gui, Add, Button, xs+10 ys+57 w65 h25 gSelectRelative hwndRelativeApplyHwnd, Apply
	Gui, Add, Button, x+5 yp wp hp gClearRelative hwndRelativeClearHwnd, Clear
	Gui, Add, Button, xs+158 ys+5 w75 h25 vUpdBt gUpdate hwndUpdateHwnd Section, Update
	Gui, Add, Button, xs ys+28 wp hp vTstBt gTest hwndTestHwnd, Test
	Gui, Add, CheckBox, xs+5 ys+56 Section Checked vFontalyzer gFontalyzerClick hwndFontHwnd, Fontalyzer
	Gui, Add, CheckBox, xs+17 ys+15 vFontGroup gFontalyzerGroupClick hwndFontGroupHwnd, Group
	Gui, Add, Progress, xs+6 ys+13 w1 h8 BackgroundBlack
	Gui, Add, Progress, xs+6 ys+21 w12 h1 BackgroundBlack
	Gui, Font, s8, Lucida Sans Typewriter
	Gui, Add, ListView, % "xm y+15 NoSortHdr Grid cNavy BackgroundC8C8C8 -LV0x10 -WantF2 vMatrix w673 h" (Min(470, 34 + 17 * Control.Count())), |x|y|w|h|Control Type|Options|Text|Index|IconNum
	ImageListID := IL_Create(3)
	LV_SetImageList(ImageListID)
	IL_Add(ImageListID, "icons\anchor_whitebg.ico", 1)
	IL_Add(ImageListID, "icons\anchor.ico", 1)
	loop, % Control.Count()
		LV_Add("Icon0",,,,,, Control[A_Index].Type, Control[A_Index].Options, Control[A_Index].Text, A_Index)
	LV_ModifyCol(1, 26)
	loop, 4
		LV_ModifyCol(A_Index + 1, 20)
	loop, 3
		LV_ModifyCol(A_Index + 5, "AutoHdr")
	LV_ModifyCol(9, 0) ; hide Index column
	LV_ModifyCol(10, 0) ; hide IconNum column
	LV_ModifyCol(1, "Icon1")
	gosub, CreateToolTips
	gosub, CreateGuiButtons
	GuiControl, Disable, UpdBt
	GuiControl, Disable, TstBt
	IniRead, WinX, settings.ini, Settings, XPos, None
	IniRead, WinY, settings.ini, Settings, YPos, None
	if !RememberPos && !Reloading
		Gui, Show,, % GuiScriptFile " - Relativity"
	else
	{
		if (WinX = "None") || (WinY = "None")
			Gui, Show,, % GuiScriptFile " - Relativity"
		else
			Gui, Show, x%WinX% y%WinY%, % GuiScriptFile " - Relativity"
	}
	GuiControl, Focus, Matrix
	LV_Modify(1, "Select Focus")
return

CreateToolTips:
	ToolTips := New GuiControlTips(MainGuiHwnd)
	ToolTips.Attach(RunHwnd, "Run the GUI script file to display the GUI`n(after saving for aligned/nudged version)")
	ToolTips.Attach(HelpBtHwnd, "Display the User Guide")
	ToolTips.Attach(XHwnd, "Align/nudge x positions of selected controls")
	ToolTips.Attach(YHwnd, "Align/nudge y positions of selected controls")
	ToolTips.Attach(WHwnd, "Align/nudge widths of selected controls")
	ToolTips.Attach(HHwnd, "Align/nudge heights of selected controls")
	ToolTips.Attach(UpBtHwnd, "Move selected row(s) up one row`n(Keyboard: Control + Up)")
	ToolTips.Attach(DownBtHwnd, "Move selected row(s) down one row`n(Keyboard: Control + Down)")
	ToolTips.Attach(TopBtHwnd, "Move selected row(s) to top`n(Keyboard: Home)")
	ToolTips.Attach(BottomBtHwnd, "Move selected row(s) to bottom`n(Keyboard: End)")
	ToolTips.Attach(GroupBtHwnd, "Group selected rows (move under top selected row)`n(Keyboard: G)")
	ToolTips.Attach(AnchorApplyHwnd, "Mark selected row(s) as a 'Section' anchor")
	ToolTips.Attach(AnchorClearHwnd, "Remove 'Section' identification from selected row(s)")
	ToolTips.Attach(XsHwnd, "Make x coordinate relative to Section control")
	ToolTips.Attach(YsHwnd, "Make y coordinate relative to Section control")
	ToolTips.Attach(XpHwnd, "Make x coordinate relative to previous control")
	ToolTips.Attach(YpHwnd, "Make y coordinate relative to previous control")
	ToolTips.Attach(WpHwnd, "Make width relative to previous control")
	ToolTips.Attach(HpHwnd, "Make height relative to previous control")
	ToolTips.Attach(RelativeApplyHwnd, "Marks selected row(s) with the relative position/size options to be applied")
	ToolTips.Attach(RelativeClearHwnd, "Clear relative position/size options from selected row(s)")
	ToolTips.Attach(UpdateHwnd, "Update control options with section labels`nand relative position/size calculations`n(automatically done before Test and Save)")
	ToolTips.Attach(TestHwnd, "Test GUI by showing changes in positions and/or sizes of`nanchor points result in expected changes in relative controls")
	ToolTips.Attach(FontHwnd, "Removes redundant font definitions from`nconsecutive controls when saving script")
	ToolTips.Attach(FontGroupHwnd, "Groups controls not related to relative positioning`nso redundant font definitions can be removed")
	ToolTips.Attach(SaveHwnd, "Save updated GUI as new script file")
	ToolTips.Attach(OpenHwnd, "Close current GUI script and open a new one")
return

CreateGuiButtons:
	GuiButtonIcon(HelpBtHwnd, "icons\help.ico", 1, "A0 L4")
	GuiButtonIcon(OpenHwnd, "icons\open.ico", 1, "A0 L4")
	GuiButtonIcon(RunHwnd, "icons\run.ico", 1, "A0 L4")
	GuiButtonIcon(XHwnd, "icons\x.ico", 1, "A0 L1 T0")
	GuiButtonIcon(YHwnd, "icons\y.ico", 1, "A0 L1 T6")
	GuiButtonIcon(WHwnd, "icons\w.ico", 1, "A0 L2")
	GuiButtonIcon(HHwnd, "icons\h.ico", 1, "A0 L1")
	GuiButtonIcon(UpBtHwnd, "icons\row_up.ico", 1)
	GuiButtonIcon(DownBtHwnd, "icons\row_down.ico", 1)
	GuiButtonIcon(TopBtHwnd, "icons\to_top.ico", 1)
	GuiButtonIcon(BottomBtHwnd, "icons\to_bottom.ico", 1)
	GuiButtonIcon(GroupBtHwnd, "icons\group.ico", 1, "A0 L2")
	GuiButtonIcon(AnchorApplyHwnd, "icons\anchor.ico", 1, "A0 L5")
	GuiButtonIcon(AnchorClearHwnd, "icons\delete.ico", 1, "A0 L4")
	GuiButtonIcon(RelativeApplyHwnd, "icons\bent_arrow.ico", 1, "A0 L2")
	GuiButtonIcon(RelativeClearHwnd, "icons\delete.ico", 1, "A0 L2")
	GuiButtonIcon(UpdateHwnd, "icons\check.ico", 1, "A0 L4")
	GuiButtonIcon(TestHwnd, "icons\checklist_gui.ico", 1, "A0 L6")
	GuiButtonIcon(SaveHwnd, "icons\save.ico", 1, "A0 L2")
return

Help:
	GuiControl, Focus, Matrix
	if WinExist("ahk_id " HelpHwnd)
		Gui, Help:Show
	else
	{
		try
		{
			Gui, Help:Destroy
			Gui, Help:+HwndHelpHwnd
			; Gui, Help:Add, Text,, HelpText
			; Gui, Help:Add, Custom, ClassRICHEDIT50W vRE hwndHRE
			Gui, Help:Add, Custom, ClassRICHEDIT50W w600 h600 vRE hwndHRE +VScroll +0x04 ; ES_MULTILINE
			Gui, Help:Show,, Relativity User Guide
			TomDoc :=  GetTomDoc(HRE)
			TomDoc.Open(A_ScriptDir "\help\Relativity User Guide.rtf", 0x01, 0)
			DllCall("SendMessage", "Ptr", HRE, "UInt", 0x044D, "Ptr", 0x02, "Ptr", 0x0800) ; EM_SETOPTIONS: ECOOP_OR, ECO_READONLY
		}
		catch
		{
			Gui, Help:Destroy
			MsgBox, 8208, Relativity Help Error, The Relativity User Guide appears to already be open in another application.  Please close the other instance in order to open the User Guide from this app.
		}
	}
return

ResetMatrix()
{
	loop, % LV_GetCount()
	{
		LV_GetText(Index, A_Index, 9)
		LV_Modify(A_Index,,,,,,,, Control[Index].Options)
	}
}

MatrixChangeCheck()
{
	GuiControlGet, SvBtnName,, SvBt
	Gui, Show,, % (Chr(0x25CF)) . (SvBtnName = "Save As" ? GuiScriptFile : OutputScriptFile) " - Relativity"
	if !MatrixChanged
	{
		ToolTip, Options will be recalculated with next Update, 215, 100
		SetTimer, ToolTipOff, -6000
		ResetMatrix()
		MatrixChanged := 1
	}
}

ToolTipOff:
	ToolTip
return

RunGuiScript()
{
	if WinExist("ahk_id " UserGuiHwnd)
		WinActivate, ahk_id %UserGuiHwnd%
	else
	{
		WinGet, activeHwnd, ID, A
		Run, % (OutputScriptFile ? OutputScriptFile : GuiScriptFile)
		WinWaitNotActive, ahk_id %activeHwnd%
		WinGet, UserGuiHwnd, ID, A
	}
	GuiControl, Focus, Matrix
}

SelectAnchor:
	RowNum := 0
	loop
	{
		RowNum := LV_GetNext(RowNum)
		if !RowNum
			break
		LV_Modify(RowNum, "Icon2",,,,,,,,,, 2)
		GuiControl, Enable, ClrBt
	}
	MatrixChangeCheck()
	EnableDisableUpdate()
	GuiControl, Focus, Matrix
return

SelectRelative:
	Gui, Submit, NoHide
	RowNum := 0
	loop
	{
		RowNum := LV_GetNext(RowNum)
		if !RowNum
			break
		LV_Modify(RowNum,,, XsBox ? "s" : (XpBox ? "p" : ""), YsBox ? "s" : (YpBox ? "p" : ""), WpBox ? "p" : "", HpBox ? "p" : "")
	}
	MatrixChangeCheck()
	EnableDisableUpdate()
	GuiControl, Focus, Matrix
return

ClearAnchor:
	RowNum := 0
	loop
	{
		RowNum := LV_GetNext(RowNum)
		if !RowNum
			break
		LV_Modify(RowNum, "Icon4",,,,,,,,,, "")
	}
	MatrixChangeCheck()
	EnableDisableUpdate()
	GuiControl, Focus, Matrix
return

ClearRelative:
	GuiControl,, XsBox, 0
	GuiControl,, YsBox, 0
	GuiControl,, XpBox, 0
	GuiControl,, YpBox, 0
	GuiControl,, WpBox, 0
	GuiControl,, HpBox, 0
	RowNum := 0
	loop
	{
		RowNum := LV_GetNext(RowNum)
		if !RowNum
			break
		LV_Modify(RowNum,,, "", "", "", "")
	}
	MatrixChangeCheck()
	EnableDisableUpdate()
	GuiControl, Focus, Matrix
return

XsBox:
	Gui, Submit, NoHide
	if XsBox
		GuiControl,, XpBox, 0
	GuiControl, Focus, Matrix
return

YsBox:
	Gui, Submit, NoHide
	if YsBox
		GuiControl,, YpBox, 0
	GuiControl, Focus, Matrix
return

XpBox:
	Gui, Submit, NoHide
	if XpBox
		GuiControl,, XsBox, 0
	GuiControl, Focus, Matrix
return

YpBox:
	Gui, Submit, NoHide
	if YpBox
		GuiControl,, YsBox, 0
	GuiControl, Focus, Matrix
return

FocusLV:
	GuiControl, Focus, Matrix
return

Update()
{
	GuiControlGet, svBtnName,, SvBt
	Menu, ScriptMenu, Enable, &Test GUI Changes
	ResetMatrix()
	Gui, Show,, % (Chr(0x25CF)) . (svBtnName = "Save As" ? GuiScriptFile : OutputScriptFile) " - Relativity"
	MatrixChanged := 0
	updatedScript := ""
	updatedScript := []
	linesAdded := xs := ys := xp := yp := wp := hp := ""
	scriptLines := StrSplit(ScriptText, "`n", "`r")
	; output header:
	loop
	{
		if (SubStr(scriptLines[A_Index], 1, 7) = "Gui Add") || (SubStr(scriptLines[A_Index], 1, 8) = "Gui Font")
			break
		updatedScript.Push(scriptLines[A_Index])
		linesAdded .= A_Index ","
		lastLineAdded := A_Index
	}
	updatedScript[1] := RegExReplace(updatedScript[1], "(; Generated by AutoGUI .*)", "$1, modified by Relativity " Version)
	; output non-located Gui line (options StatusBar, Color):
	currLine := lastLineAdded + 1
	loop
	{
		if (!RegExMatch(scriptLines[currLine], "^Gui Add, \w+, x")) && (!RegExMatch(scriptLines[currLine], "^Gui Font"))
		{
			updatedScript.Push(scriptLines[currLine])
			linesAdded .= currLine ","
		}
		currLine++
		if (Trim(scriptLines[currLine] = ""))
			break
	}
	; process and output controls and associated font definitions:
	loop, % LV_GetCount()
	{
		row := A_Index
		loop, 10
		{
			Switch A_Index
			{
				Case 2: LV_GetText(x, row, 2)
				Case 3: LV_GetText(y, row, 3)
				Case 4: LV_GetText(w, row, 4)
				Case 5: LV_GetText(h, row, 5)
				Case 6: LV_GetText(type, row, 6)
				Case 7: LV_GetText(options, row, 7)
				Case 8: LV_GetText(text, row, 8)
				Case 9: LV_GetText(index, row, 9)
				Case 10: LV_GetText(iconNum, row, 10)
			}
		}
		; get absolute numbers from original options:
		; RegExMatch(Control[index].Options, ".*\bx(\d+).*\by(\d+).*.*\bw(\d+).*\bh(\d+)", match)
		RegExMatch(Control[index].Options, "x(\d+) y(\d+).* w(\d+)(?: h(\d+))?", match) ; h is an optional capturing group
		xAbs := match1
		yAbs := match2
		wAbs := match3
		hAbs := match4

		if (x = "s" && xs = "") || (y = "s" && ys = "") || (x = "p" && xp = "") || (y = "p" && yp = "") || (w = "p" && wp = "") || (h = "p" && hp = "")
		{
			MsgBox, 8208, Relative control error, % "Row " row " contains a relative control designation that does not have an appropriate control preceding it (no section control or no previous control).`n`nPlease correct and click Update again."
			return
		}
		if (x = "s")
			options := RegExReplace(options, "\bx\d+", "xs" RelFromAbs(xs, xAbs))
		else if (x = "p")
			options := RegExReplace(options, "\bx\d+", "xp" RelFromAbs(xp, xAbs))
		if (y = "s")
			options := RegExReplace(options, "\by\d+", "ys" RelFromAbs(ys, yAbs))
		else if (y = "p")
			options := RegExReplace(options, "\by\d+", "yp" RelFromAbs(yp, yAbs))
		if (w = "p")
			options := RegExReplace(options, "\bw\d+", "wp" RelFromAbs(wp, wAbs))
		if (h = "p")
			options := RegExReplace(options, "\bh\d+", "hp" RelFromAbs(hp, hAbs))
		options := options . (iconNum = 2 ? " Section" : "")
		line := Control[index].Line
		if (RegExMatch(scriptLines[line - 1], "^Gui Font,")) && (Trim(scriptLines[line + 1]) = "Gui Font")
		{
			updatedScript.Push(scriptLines[line - 1])
			updatedScript.Push("Gui Add, " type ", " Options . (text ? ", " : "") . text)
			updatedScript.Push(ScriptLines[line + 1])
			linesAdded .= (line - 1) "," line "," (line + 1) ","
		}
		else
		{
			updatedScript.Push("Gui Add, " type ", " options . (text ? ", " : "") . text)
			linesAdded .= line ","
		}
		; update the Options column in the ListView:
		LV_Modify(A_Index,,,,,,,, options)
		; set baseline values for next section or previous relative options
		if (iconNum = 2) ; section
		{
			xs := xAbs
			ys := yAbs
		}
		xp := xAbs
		yp := yAbs
		wp := wAbs
		hp := hAbs
	}
	; resize the columns:
	loop, 3
		LV_ModifyCol(A_Index + 5, "AutoHdr")	

	; output the remaining lines:
	linesAdded := SubStr(linesAdded, 1, -1) ; removes the last comma
	loop, % scriptLines.Count()
		if A_Index not in %linesAdded%
			updatedScript.Push(scriptLines[A_Index])
	
	OutputScriptText := ""
	loop, % updatedScript.Count()
		OutputScriptText .= updatedScript[A_Index] "`n"

	GuiControl, Enable, TstBt
	GuiControl, Focus, Matrix
}

RelFromAbs(b, a) ; baseline, absolute
{
	return a - b > 0 ? "+" a - b : (a - b < 0 ? a - b : "")
}

Test()
{
	Testing := 1
	newControl := []
	sectionAnchor := []
	virtualAnchor := []
	Update()
	Gui, Cancel
	; parse updated script:
	loop, Parse, OutputScriptText, `n
		if RegExMatch(A_LoopField, "^Gui Add, \w+, x")
			newControl.Push(new ControlDef(A_LoopField, A_Index))
	; identify anchors:
	loop, % newControl.Count()
		if InStr(newControl[A_Index].Options, "Section")
			sectionAnchor.Push(A_Index)
	loop, % newControl.Count()
	{
		row := A_Index
		loop, 4
			f%A_Index% := 0
		loop, 4
			if (!InStr(newControl[row].Options, ["xp", "yp", "wp", "hp"][A_Index]) && InStr(newControl[row + 1].Options, ["xp", "yp", "wp", "hp"][A_Index]))
				f%A_Index% := 1
		if (f1 || f2 || f3 || f4)
			virtualAnchor.Push({index: row, xy: f1 || f2, w: f3, h: f4})
	}
	ToolbarMsg := (InStr(ScriptText, "hToolbar := CreateToolbar()") ? "`n`nThe toolbar has been`nremoved for this test" : "")
	SplashMessage("The GUI script has " sectionAnchor.Count() " section" (sectionAnchor.Count() = 1 ? "" : "s") . "`nand " virtualAnchor.Count() " 'previous' test point" (virtualAnchor.Count() = 1 ? "" : "s") . ToolbarMsg "`n`nPress Esc to quit testing", ToolbarMsg ? 5 : 3, 260, ToolbarMsg ? 160 : 140, 1)
	if !Testing
	{
		QuitTesting()
		return
	}
	; test sections:
	Sleep, 500
	loop, % sectionAnchor.Count()
	{
		SplashMessage("Testing section " A_Index "`n`nSection " A_Index " will move`nfor a few seconds",,,, 1)
		if !Testing
		{
			QuitTesting()
			return
		}
		RunTestGuis(ScriptText, AltTestScript(newControl[sectionAnchor[A_Index]].Line, "xy"))
	}
	; test 'previous' relative controls:
	loop, % virtualAnchor.Count()
	{
		if virtualAnchor[A_Index].xy
		{
			SplashMessage("Testing position of test point " A_Index "`n`nTest point " A_Index " will move`nfor a few seconds",, 230,, 1)
			if !Testing
			{
				QuitTesting()
				return
			}
			RunTestGuis(ScriptText, AltTestScript(newControl[virtualAnchor[A_Index].index].Line, "xy"))
		}
		if virtualAnchor[A_Index].w
		{
			SplashMessage("Testing width of test point " A_Index "`n`nTest point " A_Index " will change width`nfor a few seconds",, 230,, 1)
			if !Testing
			{
				QuitTesting()
				return
			}
			RunTestGuis(ScriptText, AltTestScript(newControl[virtualAnchor[A_Index].index].Line, "w"))
		}
		if virtualAnchor[A_Index].h
		{
			SplashMessage("Testing height of test point " A_Index "`n`nTest point " A_Index " will change height`nfor a few seconds",, 230,, 1)
			if !Testing
			{
				QuitTesting()
				return
			}
			RunTestGuis(ScriptText, AltTestScript(newControl[virtualAnchor[A_Index].index].Line, "h"))
		}
	}
	Gui, Show
	GuiControl, Focus, Matrix
}
return

FontalyzerClick:
	Gui, Submit, NoHide
	GuiControl, % (Fontalyzer ? "Enable" : "Disable"), FontGroup
	Menu, ScriptMenu, % (Fontalyzer ? "Check" : "Uncheck"), &Fontalyzer
	Menu, ScriptMenu, % (Fontalyzer ? "Enable" : "Disable"), &Fontalyzer Grouping
	GuiControl, Focus, Matrix
return

FontalyzerGroupClick:
	Gui, Submit, NoHide
	Menu, ScriptMenu, % (FontGroup ? "Check" : "Uncheck"), &Fontalyzer Grouping
	GuiControl, Focus, Matrix
return

FontalyzerMenuSelect:
	Gui, Submit, NoHide
	GuiControl,, Fontalyzer, % (Fontalyzer ? 0 : 1)
	GuiControl, % (Fontalyzer ? "Disable" : "Enable"), FontGroup
	Menu, ScriptMenu, % (Fontalyzer ? "Uncheck" : "Check"), &Fontalyzer
	Menu, ScriptMenu, % (Fontalyzer ? "Disable" : "Enable"), &Fontalyzer Grouping
return

FontalyzerGroupMenuSelect:
	Gui, Submit, NoHide
	GuiControl,, FontGroup, % !FontGroup
	Menu, ScriptMenu, % (FontGroup ? "Uncheck" : "Check"), &Fontalyzer Grouping
return

WelcomeMenuSelect:
	ShowWelcome := !ShowWelcome
	Menu, OptionsMenu, % (ShowWelcome ? "Check" : "Uncheck"), Show &Welcome Screen
	IniWrite, % ShowWelcome, settings.ini, Settings, ShowWelcome
return

RememberPosMenuSelect:
	RememberPos := !RememberPos
	Menu, OptionsMenu, % (RememberPos ? "Check" : "Uncheck"), &Remember Window Position
	IniWrite, % RememberPos, settings.ini, Settings, RememberPos
return

Save:
	Update()
	GuiControlGet, SvBtnName,, SvBt
	SuggestedFile := RegExReplace(GuiScriptFile, ".ahk$", "_rel.ahk")
	if (SvBtnName = "Save As")
	{
		FileSelectFile, OutputScriptFile, S18, % SuggestedFile, Select name and location for GUI script output, *.ahk
		if !OutputScriptFile
			return
		GuiControl,, SvBt, Save
		GuiButtonIcon(SaveHwnd, "shell32.dll", 259, "A0 L6")
		ToolTips.Update(SaveHwnd, "Save updated GUI to the new script file")
		Menu, FileMenu, Rename, &Save As..., &Save
	}
	FileDelete, % OutputScriptFile
	FileAppend, % (Fontalyzer ? Fontalyze(OutputScriptText, FontGroup) : OutputScriptText), % OutputScriptFile
	Gui, Show,, % OutputScriptFile " - Relativity"
	GuiControl, Focus, Matrix
return

IntroGui()
{
	Gui, Intro:-MinimizeBox -MaximizeBox +HwndintroHwnd
	Gui, Intro:Color, 8090A8
	Gui, Intro:Add, Progress, x0 y0 w500 h55 Section BackgroundE4E4E4
	Gui, Intro:Font, s9 cNavy, Lucida Sans
	Gui, Intro:Add, Text, xs+30 ys+10 BackgroundTrans, Adds relative positioning to controls created by AutoGUI's GUI designer.
	Gui, Intro:Add, Text, xp y+5 BackgroundTrans c800000, Must use a GUI script file that is unmodified from the AutoGUI output.
	Gui, Intro:Add, Progress, x0 y55 w500 h135 Section BackgroundWhite
	Gui, Intro:Add, Text, x30 ys+10 BackgroundTrans, Functionality:
	Gui, Intro:Add, Progress, xm+20 y+10 w5 h5 BackgroundBlack
	Gui, Intro:Add, Text, x+10 yp-5 BackgroundTrans, Positions and sizes controls relative to previous or section controls
	Gui, Intro:Add, Progress, xm+20 y+10 w5 h5 BackgroundBlack
	Gui, Intro:Add, Text, x+10 yp-5 BackgroundTrans, Performs the math for converting absolute controls to relative controls
	Gui, Intro:Add, Progress, xm+20 y+10 w5 h5 BackgroundBlack
	Gui, Intro:Add, Text, x+10 yp-5 BackgroundTrans, Reorders script lines (with font specifications) as necessary
	Gui, Intro:Add, Progress, xm+20 y+10 w5 h5 BackgroundBlack
	Gui, Intro:Add, Text, x+10 yp-5 BackgroundTrans, Optionally removes redundant consecutive font definitions
	Gui, Intro:Add, Progress, xm+20 y+10 w5 h5 BackgroundBlack
	Gui, Intro:Add, Text, x+10 yp-5 BackgroundTrans, Test feature demonstrates that controls were grouped as intended
	Gui, Intro:Add, Button, x25 yp+35 w140 Default gSelectScriptBtn, Select GUI Script
	Gui, Intro:Add, Button, x+15 wp gHelp, View Help
	Gui, Intro:Add, Button, x+15 wp gIntroExitBtn, Exit
	Gui, Intro:Show, w500, Relativity
	WinWaitClose, ahk_id %introHwnd%
}
return

About:
	GuiControl, Focus, Matrix
	MsgBox, 8256, About Relativity, Relativity`, v%Version%`n`nCreated by boiler`, a member of the AutoHotkey forums.  Provided for non-commercial use without restriction.`n`nThe author would like to acknowledge the users Alguimist`, lexikos`, HotKeyIt`, just me`, FanaticGuru`, and burque505 for their tools and functions as cited in the source code.
return

SelectScriptBtn:
	Gui, Intro:Cancel
return

GuiSize:
	if (A_EventInfo != 0)
		return
	GuiControl, Move, Matrix, % "w" A_GuiWidth - 20 " h" A_GuiHeight - 111
return

ExitRelativity:
GuiClose:
	WinGetTitle, GuiTitle, ahk_id %MainGuiHwnd%
	if (Asc(SubStr(GuiTitle, 1, 1)) = 0x25CF)
	{
		MsgBox, 8260, Unsaved changes, % "There are unsaved changes.  Are you sure you want to exit the program without saving?"
		IfMsgBox, No
		{
			GuiControl, Focus, Matrix
			return
		}
	}
	WinGetPos, WinX, WinY,,, ahk_id %MainGuiHwnd%
	if (WinX != "")  && (WinY != "") {
		IniWrite, % WinX, settings.ini, Settings, XPos
		IniWrite, % WinY, settings.ini, Settings, YPos
	}
	WinClose, ahk_id %UserGuiHwnd%
ExitApp


OpenFile:
	WinGetTitle, GuiTitle, ahk_id %MainGuiHwnd%
	if (Asc(SubStr(GuiTitle, 1, 1)) = 0x25CF)
	{
		MsgBox, 8260, Unsaved changes, % "There are unsaved changes.  Are you sure you want to open a new file without saving the existing file?"
		IfMsgBox, No
		{
			GuiControl, Focus, Matrix
			return
		}
	}
	WinGetPos, WinX, WinY,,, ahk_id %MainGuiHwnd%
	if (WinX != "")  && (WinY != "") {
		IniWrite, % WinX, settings.ini, Settings, XPos
		IniWrite, % WinY, settings.ini, Settings, YPos
	}

	Gui, +OwnDialogs
	loop
	{
		FileSelectFile, ThisScriptFile, 1,, % "Select GUI script file", *.ahk
		if !ThisScriptFile
		{
			SplashMessage("No file selected`n`nReturning to prior file")
			return
		}
		FileRead, ThisScriptText, % ThisScriptFile
	} until ScriptIsValid(ThisScriptText)

	IniWrite, 1, settings.ini, Status, Reloading
	IniWrite, % ThisScriptFile, settings.ini, Status, ReloadFile

	WinClose, ahk_id %UserGuiHwnd%
Reload

IntroGuiClose:
IntroExitBtn:
ExitApp

#If WinActive("ahk_id " MainGuiHwnd) && !Busy
UpBtn:
^Up::
	Busy := 1
	SelectedRows := GetSelectedRows()
	if (SelectedRows[1] > 1)
	{
		MatrixChangeCheck()
		loop, % SelectedRows.Count()
			MoveRow(SelectedRows[A_Index], SelectedRows[A_Index] - 1)
		LV_Modify(SelectedRows[1] - 1, "Vis Focus")
	}
	GuiControl, Focus, Matrix
	Busy := 0
return

DownBtn:
^Down::
	Busy := 1
	SelectedRows := GetSelectedRows()
	if (SelectedRows[1] != 0) && (SelectedRows[SelectedRows.Count()] < Control.Count())
	{
		MatrixChangeCheck()
		loop, % SelectedRows.Count()
		{
			ThisIndex := SelectedRows.Count() - A_Index + 1
			MoveRow(SelectedRows[ThisIndex], SelectedRows[ThisIndex] + 1)
		}
		LV_Modify(SelectedRows[ThisIndex] + 1, "Vis Focus")
	}
	GuiControl, Focus, Matrix
	Busy := 0
return

TopBtn:
Home::
	Busy := 1
	SelectedRows := GetSelectedRows()
	if (SelectedRows[1] != 0)
	{
		MatrixChangeCheck()
		loop, % SelectedRows.Count()
			MoveRow(SelectedRows[A_Index], A_Index)
		LV_Modify(1, "Vis Focus")
	}
	GuiControl, Focus, Matrix
	Busy := 0
return

BottomBtn:
End::
	Busy := 1
	SelectedRows := GetSelectedRows()
	if (SelectedRows[1] != 0)
	{
		MatrixChangeCheck()
		; move last one first:
		loop, 10
			LV_GetText(t%A_Index%, SelectedRows[SelectedRows.Count()], A_Index)
		LV_Delete(SelectedRows[SelectedRows.Count()])
		LV_Add("Select Icon" . (t10 ? t10 : 5), t1, t2, t3, t4, t5, t6, t7, t8, t9, t10)

		; insert the rest before that one (from the bottom up)
		loop, % SelectedRows.Count() - 1
		{
			ThisIndex := SelectedRows.Count() - A_Index
			loop, 10
				LV_GetText(t%A_Index%, SelectedRows[ThisIndex], A_Index)
			LV_Delete(SelectedRows[ThisIndex])
			LV_Insert(Control.Count() - A_Index, "Select Icon" . (t10 ? t10 : 5), t1, t2, t3, t4, t5, t6, t7, t8, t9, t10)
		}
		LV_Modify(Control.Count(), "Vis Focus")
	}
	GuiControl, Focus, Matrix
	Busy := 0
return

GroupBtn:
g::
	Busy := 1
	SelectedRows := GetSelectedRows()
	if (SelectedRows[1] != 0)
	{
		MatrixChangeCheck()
		loop, % SelectedRows.Count() - 1
			MoveRow(SelectedRows[A_Index + 1], SelectedRows[1] + A_Index)
		LV_Modify(SelectedRows[1], "Vis Focus")
	}
	GuiControl, Focus, Matrix
	Busy := 0
return

#If WinActive("ahk_id " MainGuiHwnd)
^Up::
^Down::
Home::
End::
g::
return ; to null these keys while busy

#If Testing = 1
Esc::Testing := 0
#If

x:
y:
w:
h:
	Align(A_ThisLabel)
return

Align(element)
{
	AlignIndex := ""
	AlignIndex := []
	opt := []
	AlignDelta := 0
	AlignSelectedRows := GetSelectedRows()
	for each, rowNum in AlignSelectedRows
	{
		LV_GetText(thisOpt, rowNum, 7)
		LV_GetText(thisIndex, rowNum, 9)
		opt.Push(thisOpt)
		AlignIndex.Push(thisIndex)
	}
	RegExMatch(opt[1], "(?<=" element ")[sp]", type)
	AlignBase := element . type
	loop, % opt.Count() - 1
	{
		RegExMatch(opt[A_Index + 1], element "[sp]{0,1}", match)
		if (match != AlignBase)
		{
			SplashMessage("All controls must have the same type`nof positioning or sizing, such as " AlignBase . (StrLen(AlignBase) = 1 ? "#" : ""),, 260, 80)
			return
		}
	}
	RegExMatch(opt[1], AlignBase "\S*", startAlignText)
	RegExMatch(startAlignText, "\d+", StartAlignValue)
	StartAlignValue := StartAlignValue = "" ? 0 : StartAlignValue
	CurrAlignVal := StartAlignValue + AlignDelta
	WinGetPos, wx, wy
	Gui, 1:+Disabled
	Gui, Align:+Owner1 -Caption
	Gui, Align:Color, 6080A0
	Gui, Align:Font, s12, Lucida Sans Typewriter
	Gui, Align:Add, Progress, x0 y0 w109 h1 BackgroundA0C0F0
	Gui, Align:Add, Progress, x0 y0 w1 h96 BackgroundA0C0F0
	Gui, Align:Add, Progress, x108 y1 w1 h96 Background4060A0
	Gui, Align:Add, Progress, x1 y95 w109 h1 Background4060A0
	Gui, Align:Add, Progress, x5 y5 w81 h30 BackgroundFDFDD9
	Gui, Align:Add, Text, xp yp wp hp Section Center BackgroundTrans +0x200 vAlignText, % startAlignText
	Gui, Align:Font
	Gui, Align:Add, UpDown, x+2 yp h30 Range-5000-5000 vAlignDelta gAlignUpDown -16
	Gui, Align:Add, Progress, xs ys w80 h1 Background4060A0
	Gui, Align:Add, Progress, xs ys w1 h30 Background4060A0
	Gui, Align:Add, Progress, xs+80 ys w1 h30 BackgroundA0C0F0
	Gui, Align:Add, Progress, xs ys+29 w80 h1 BackgroundA0C0F0
	Gui, Align:Add, Button, xs y+5 w100 gAlignOK, OK
	Gui, Align:Add, Button, xs y+5 wp hp gAlignCancel, Cancel
	Gui, Align:Show, % "x" wx + 650 " y" wy + 170 " w109 h96"
}

AlignUpDown:
	Gui, Align:Submit, NoHide
	CurrAlignVal := StartAlignValue + AlignDelta
	if (InStr(AlignBase, "s") || InStr(AlignBase, "p"))
		GuiControl,, AlignText, % AlignBase . (CurrAlignVal = 0 ? "" : (CurrAlignVal < 0 ? CurrAlignVal : "+" CurrAlignVal))
	else
		GuiControl,, AlignText, % AlignBase . CurrAlignVal
return

AlignOK()
{
	Gui, 1:Default
	element := SubStr(AlignBase, 1, 1)
	opt := Control[AlignIndex[1]].Options
	RegExMatch(opt, "(?<=" element ")\d+", abs)
	replace := element . (abs + AlignDelta)
	; change Options in the Control array:
	for each, index in AlignIndex
	{
		opt := Control[index].Options
		Control[index].Options := RegExReplace(opt, element "\d+", replace,, 1)
	}
	; change the controls displayed in the matrix:
	for each, rowNum in AlignSelectedRows
	{
		LV_GetText(opt, rowNum, 7)
		if (InStr(AlignBase, "s") || InStr(AlignBase, "p"))
			replace := AlignBase . (CurrAlignVal = 0 ? "" : (CurrAlignVal < 0 ? CurrAlignVal : "+" CurrAlignVal))
		else
			replace := AlignBase . CurrAlignVal
		opt := RegExReplace(opt, AlignBase "[+-]{0,1}\d*", replace,, 1)
		LV_Modify(rowNum,,,,,,,, opt)
	}
	; change the lines in the baseline loaded script text:
	abs := ""
	lines := StrSplit(ScriptText, "`n", "`r")
	for each, index in AlignIndex
	{
		line := lines[Control[index].line]
		if !abs
		{
			RegExMatch(line, ".*?,.*?, \K[^,]+", options)
			RegExMatch(options, "(?<=" element ")\d+", abs)
			replace := abs + AlignDelta
		}
		lines[Control[index].line] := RegExReplace(line, element . abs, element . replace,, 1)
	}
	ScriptText := ""
	for each, line in lines
		ScriptText .= line "`n"
	Gui, Align:Destroy
	Gui, 1:-Disabled
	WinActivate, ahk_id %MainGuiHwnd%
	GuiControl, Focus, Matrix
	GuiControlGet, SvBtnName,, SvBt
	Gui, Show,, % (Chr(0x25CF)) . (SvBtnName = "Save As" ? GuiScriptFile : OutputScriptFile) " - Relativity"
}

AlignGuiClose:
AlignCancel:
	Gui, Align:Destroy
	Gui, 1:-Disabled
	WinActivate, ahk_id %MainGuiHwnd%
	GuiControl, Focus, Matrix
return
